import pygetwindow as gw
import time
import sys
import os
import win32gui
import win32con
import win32process
import ctypes
from ctypes import wintypes
import psutil

try:
    from . config import CONFIG
except:
    from config import CONFIG

config = CONFIG()

def set_window_always_on_top(hwnd):
    """Set a window to always be on top."""
    win32gui.SetWindowPos(
        hwnd,
        win32con.HWND_TOPMOST,
        0, 0, 0, 0,
        win32con.SWP_NOMOVE | win32con.SWP_NOSIZE
    )

def unset_window_always_on_top(hwnd):
    """Unset a window's always-on-top property."""
    win32gui.SetWindowPos(
        hwnd,
        win32con.HWND_NOTOPMOST,
        0, 0, 0, 0,
        win32con.SWP_NOMOVE | win32con.SWP_NOSIZE
    )

def restore_and_focus_window(hwnd):
    """Restore and bring a window to the foreground."""
    if win32gui.IsIconic(hwnd):  # Check if the window is minimized
        win32gui.ShowWindow(hwnd, win32con.SW_RESTORE)  # Restore the window
    # Attempt to bring the window to the foreground
    if not win32gui.SetForegroundWindow(hwnd):
        attach_thread_input_and_focus(hwnd)

def attach_thread_input_and_focus(hwnd):
    """Attach input threads to set a window as foreground."""
    current_thread_id = win32process.GetCurrentThreadId()
    foreground_window = win32gui.GetForegroundWindow()
    if foreground_window:
        foreground_thread_id, _ = win32process.GetWindowThreadProcessId(foreground_window)
        if ctypes.windll.user32.AttachThreadInput(current_thread_id, foreground_thread_id, True):
            win32gui.SetForegroundWindow(hwnd)
            ctypes.windll.user32.AttachThreadInput(current_thread_id, foreground_thread_id, False)

# Get the current active window before making changes
def get_active_window():
    hwnd = win32gui.GetForegroundWindow()
    return hwnd

# Find Windows Terminal window
def find_terminal_window():
    for window in gw.getWindowsWithTitle('Windows Terminal'):
        if 'Windows Terminal' in window.title:
            return window
    return None

def get_window_title(hwnd):
    # Define buffer size
    length = 256  # Maximum length of a window title
    buffer = ctypes.create_unicode_buffer(length)
    # Call GetWindowTextW to fetch the title
    ctypes.windll.user32.GetWindowTextW(hwnd, buffer, length)
    return buffer.value

def get_parent_window_handle():
    find_pid = None
    # Callback function for EnumWindows
    def enum_windows_callback(hwnd, lParam):
        # Get the process ID associated with this window
        process_id = wintypes.DWORD()
        ctypes.windll.user32.GetWindowThreadProcessId(hwnd, ctypes.byref(process_id))
        # if process_id.value == parent_pid:
        if process_id.value == find_pid:
            hwnd_list.append(hwnd)
        return True

    hwnd_list = []

    # Get the parent process ID
    current_process = psutil.Process(os.getpid())
    # print(f"current_process: {current_process}")
    # parent_pid = current_process.ppid()  # Get parent process ID
    parents = current_process.parents()
    # print(f"parent_pid: {parents}")

    accepts = config.ACCEPTS or config._data_default.get('ACCEPTS') or ['WindowsTerminal.exe', 'cmd.exe', 'python.exe']

    for p in parents:
        # print(f"p.name(): {p.name()}")
        if p.name() in accepts:
            find_pid = p.pid
    
    # print(f"find_pid: {find_pid}")

    if find_pid:
        # Call EnumWindows to enumerate all top-level windows
        EnumWindows = ctypes.windll.user32.EnumWindows
        EnumWindowsProc = ctypes.WINFUNCTYPE(wintypes.BOOL, wintypes.HWND, wintypes.LPARAM)
        EnumWindows(EnumWindowsProc(enum_windows_callback), 0)

    # print(f"hwnd_list: {hwnd_list}")
    # Return the first matching window handle, if found
    return hwnd_list[0] if hwnd_list else None

def set():
    active_window = get_active_window()
    # print(f"active_window: {active_window}")
    title = get_window_title(active_window)
    # print(f"title: {title}")
    active_terminal = get_parent_window_handle()
    # print(f"active_terminal: {active_terminal}")
    title_active_terminal = get_window_title(active_terminal)
    # print(f"title_active_terminal: {title_active_terminal}")
    # print("Setting Windows Terminal to always on top and bringing it to foreground.")
    set_window_always_on_top(active_terminal)
    # win32gui.SetForegroundWindow(active_terminal)  # Bring to foreground
    time.sleep(
        config.SLEEP or config._data_default.get('SLEEP') or 7
    )
    unset_window_always_on_top(active_terminal)
    win32gui.SetForegroundWindow(active_window)  # Bring to foreground
    
    # sys.exit()
    # Find Windows Terminal window
    # terminal_window = find_terminal_window()
    # if terminal_window is None:
    #     print("Windows Terminal is not open.")
    # else:
    #     # Set Windows Terminal always on top and bring it to the foreground
    #     print("Setting Windows Terminal to always on top and bringing it to foreground.")
    #     set_window_always_on_top(terminal_window._hWnd)
    #     win32gui.SetForegroundWindow(terminal_window._hWnd)  # Bring to foreground

    #     # Wait for 10 seconds
    #     time.sleep(10)

    #     # Revert Windows Terminal always-on-top
    #     print("Reverting Windows Terminal to normal.")
    #     unset_window_always_on_top(terminal_window._hWnd)

    #     # Restore focus to the previously active window
    #     if active_window and win32gui.IsWindow(active_window):
    #         try:
    #             restore_and_focus_window(active_window)
    #         except Exception as e:
    #             print(f"Error restoring previous window: {e}")



import logging
from logging.handlers import SysLogHandler
# import os
import socket
import traceback
import shutil
import pickle
import argparse

from rich.console import Console
from rich.traceback import Traceback #, Trace
# Initialize Rich Console
console = Console()

try:
    from . config import CONFIG
except:
    from config import CONFIG

class CTraceback:

    def __init__(self, exc_type, exc_value, tb):
        self.config = CONFIG()
        # Create syslog handler (UDP for performance)
        self.syslog_handler = SysLogHandler(address=(self.config.SYSLOG_SERVER, int(self.config.SYSLOG_PORT) if self.config.SYSLOG_PORT else 514), socktype=socket.SOCK_DGRAM)
        self.syslog_handler.setLevel(logging.ERROR)
        self.syslog_formatter = logging.Formatter('%(message)s')
        self.syslog_handler.setFormatter(self.syslog_formatter)

        # Create logger for syslog only
        self.syslog_logger = logging.getLogger("SyslogOnly")
        self.syslog_logger.addHandler(self.syslog_handler)
        self.syslog_logger.setLevel(logging.ERROR)
        self.syslog_logger.propagate = False

        # Add a file handler for traceback.log
        self.file_handler = logging.FileHandler(self.config.LOG_FILE or self.config._data_default.get('LOG_FILE'))
        self.file_handler.setLevel(logging.ERROR)
        self.file_formatter = logging.Formatter('%(asctime)s - %(message)s')
        self.file_handler.setFormatter(self.file_formatter)

        # Create logger for file only
        self.file_logger = logging.getLogger("FileOnly")
        self.file_logger.addHandler(self.file_handler)
        self.file_logger.setLevel(logging.ERROR)
        self.file_logger.propagate = False

        self.custom_exception_handler(exc_type, exc_value, tb)
 
    def custom_exception_handler(self, exc_type, exc_value, tb):
        # Generate plain-text traceback once
        plain_traceback = ''.join(traceback.format_exception(exc_type, exc_value, tb))

        tb_renderable = Traceback.from_exception(exc_type, exc_value, tb, show_locals=True if self.config.SHOW_LOCAL in ["1", 'true', 'True'] else False, width=shutil.get_terminal_size()[0], theme = self.config.THEME or 'fruity')
        console.print(tb_renderable)

        # Log to syslog only
        # syslog_message = f"{DEFAULT_TAG}: Complete Traceback:\n{plain_traceback.strip()}"
        syslog_message = f"{self.config.DEFAULT_TAG}: {plain_traceback.strip()}"
        self.syslog_logger.error(syslog_message)

        # Log to file only
        log_message = f"Complete Traceback:\n{plain_traceback.strip()}"
        self.file_logger.error(log_message)

        if self.config.TRACEBACK_ACTIVE == 1:
            # Extract traceback details as a string
            tb_details = "".join(traceback.format_tb(tb))

            # Serialize the exception data
            serialized_data = pickle.dumps((exc_type.__name__, str(exc_value), tb_details))

            # Send the data to the server
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as client:
                client.connect((self.config.TRACEBACK_SERVER, self.config.TRACEBACK_PORT))
                client.sendall(serialized_data)

    @classmethod
    def usage(self):
        import sys

        parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
        server_subparser = parser.add_subparsers(dest = "runas", help = "Server arguments")
        
        serve_args = server_subparser.add_parser('serve', help = "Run as server")
        serve_args.add_argument('-b', '--host', default = "127.0.0.1", type=str, help = 'listen on ip/host')
        serve_args.add_argument('-p', '--port', default = 7000, type = int, help = "listen on port number (TCP)")

        parser.add_argument('-t', '--test', action='store_true', help = "Test exception")

        args = parser.parse_args()

        if len(sys.argv) == 1:
            parser.print_help()
        else:
            if args.test:
                import sys
                sys.excepthook = CTraceback


                # Example to trigger an exception
                def example_error():
                    raise ValueError("This is a test error for traceback handling!")

                example_error()
            elif args.runas == 'serve':
                if args.host or args.port != 7000:
                    try:
                        from . import server
                    except:
                        import server
                    server.start_server(args.host, args.port)
                else:
                    parser.print_help()
            else:
                parser.print_help()

# Example usage
if __name__ == "__main__":
    CTraceback.usage()


import socket
import pickle
import traceback
from rich.console import Console
from rich.text import Text
import shutil
import datetime
import sys
import signal

try:
    from . config import CONFIG
except:
    from config import CONFIG

if sys.platform == 'win32':
    try:
        from . import on_top
    except:
        import on_top

config = CONFIG()

# Server configuration
HOST = config.TRACEBACK_SERVER or config._data_default.get('TRACEBACK_SERVER') or '127.0.0.1'
PORT = int(str(config.TRACEBACK_PORT or config._data_default.get('TRACEBACK_PORT') or 7000))

# Rich console for colorized output
console = Console()

# Function to format and print traceback with colors
def print_traceback(exc_type, exc_value, tb_details):
    terminal_width = shutil.get_terminal_size()[0]

    # Timestamp
    timestamp = datetime.datetime.now().strftime("[bold #FF00FF]%Y[/]-[bold #0055FF]%m[/]-[bold #FF55FF]%d[/] [bold #FFFF00]%H[/]:[bold #FF5500]%M[/]:[bold #AAAAFF]%S[/].[bold #00FF00]%f[/]")
    console.print(f"[bold]{timestamp}[/bold] - ", end='')

    # Format traceback parts with colors
    type_text = Text(str(exc_type), style="white on red blink")
    value_text = Text(str(exc_value), style="black on #FFFF00")
    # tb_text = Text("".join(traceback.format_tb(tb)), style="green")
    tb_text = Text(tb_details, style="#00FFFF")

    # Print the traceback parts
    console.print(type_text, end = '')
    console.print(": ", end = '')
    console.print(value_text)
    console.print(tb_text)

    # Separator line
    console.print("-" * terminal_width)

# Server to listen for traceback data
def start_server(host = None, port = None):
    global server_socket
    
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as server:
            console.print(f"[bold #FFAA00]Server is listening on[/] [bold #00FFFF]{host or HOST}[/]:[bold #FF55FF]{port or PORT}[/]")

            server_socket = server
            server.bind((host or HOST, int(port or PORT)))
            server.listen()
            
            while True:
                conn, addr = server.accept()
                with conn:
                    console.print(f"[blue]Connected by {addr}[/blue]")

                    # Receive serialized data
                    data = b""
                    while True:
                        packet = conn.recv(4096)
                        if not packet:
                            break
                        data += packet
                        if config.ON_TOP == 1 and sys.platform == 'win32':
                            on_top.set()

                    # Deserialize data
                    exc_type, exc_value, tb_details = pickle.loads(data)

                    # Print the traceback
                print_traceback(exc_type, exc_value, tb_details)
    except KeyboardInterrupt:
        server_socket.close()
        sys.exit()

def handle_exit_signal(signum, frame):
    """Handle termination signals."""
    console.print("\n[bold #FF00FF]Shutting down server...[/]")
    if server_socket:
        server_socket.close()
    sys.exit(0)

# Register signal handlers
signal.signal(signal.SIGINT, handle_exit_signal)  # Handle Ctrl+C (SIGINT)
signal.signal(signal.SIGTERM, handle_exit_signal)  # Handle termination signal

if __name__ == "__main__":
    start_server()



from pathlib import Path
from configset import configset
import json
from pydebugger.debug import debug
import os

class CONFIG:
    _config_file = Path.cwd() / 'traceback.json' if (Path.cwd() / 'traceback.json').is_file() else '' or Path(__file__).parent / "traceback.json"
    _config_ini_file = str(Path.cwd() / 'traceback.ini') if (Path.cwd() / 'traceback.ini').is_file() else "" or str(Path(__file__).parent / "traceback.ini")
    config = configset(_config_ini_file)

    _data = {
        # Remote Syslog configuration
        'SYSLOG_SERVER' : config.get_config('syslog', 'host') or '127.0.0.1',
        'SYSLOG_PORT' : str(config.get_config('syslog', 'port')) if config.get_config('syslog', 'port') else '' or '514',
        'TRACEBACK_SERVER' : config.get_config('server', 'host') or '127.0.0.1',
        'TRACEBACK_PORT' : str(config.get_config('server', 'port')) if config.get_config('server', 'port') else '' or 7000,
        'TRACEBACK_ACTIVE' : config.get_config('server', 'active') if config.get_config('server', 'active') in [1,0] else 1,
        'DEFAULT_TAG' : config.get_config('tag', 'name') or "TracebackLogger",
        'SHOW_LOCAL'  : str(config.get_config('rich', 'show_local')) if config.get_config('rich', 'show_local') else '' or "1",
        'THEME'  : config.get_config('rich', 'theme') if config.get_config('rich', 'theme') else '' or "fruity",
        # Configure traceback logging
        'LOG_FILE' : config.get_config('file', 'name') or os.path.join(str(Path(__file__).parent), "traceback.log"),
        'ACCEPTS' : config.get_config_as_list('on_top', 'accepts') or ['WindowsTerminal.exe', 'cmd.exe', 'python.exe'],
        'ON_TOP' : int(config.get_config('on_top', 'active')) if config.get_config('on_top', 'active') else '' or 1,
        'SLEEP' : int(config.get_config('on_top', 'sleep')) if config.get_config('on_top', 'sleep') else '' or 7,
    }

    debug(_data = _data)

    _data_default = _data

    def __init__(self):
        # Load existing configuration if the file exists
        if self._config_file.exists():
            with open(self._config_file, "r") as f:
                self._data = json.load(f)

    def __getattr__(self, name):
        # Retrieve a value from the configuration data
        if name in self._data:
            return self._data[name]
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")

    def __setattr__(self, name, value):
        if name in {"_config_file", "_data"}:  # Allow setting internal attributes
            super().__setattr__(name, value)
        else:
            # Update the configuration data and save to the file
            self._data[name] = value
            with open(self._config_file, "w") as f:
                json.dump(self._data, f, indent=4)




import io
from setuptools import setup, find_packages
import sys

with io.open("README.md", "rt", encoding="utf8") as f:
    readme = f.read()

version = {}
with open("__version__.py") as fp:
    exec(fp.read(), version)
version = version['version']

if sys.platform == 'win32':
    install_requires = [
        'pygetwindow',
        'argparse',
        'configset', 
        'rich',
        'pydebugger',
        'ctype',
        'pywin32',  
        'psutil'
    ]
else:
    install_requires = [
        'argparse',
        'configset', 
        'rich',
        'pydebugger',
    ]

setup(
    name="ctraceback",
    version=version,
    url="https://github.com/cumulus13/ctraceback",
    project_urls={
        "Documentation": "https://github.com/cumulus13/ctraceback",
        "Code": "https://github.com/cumulus13/ctraceback",
    },
    license="GPL",
    author="Hadi Cahyadi LD",
    author_email="cumulus13@gmail.com",
    maintainer="cumulus13 Team",
    maintainer_email="cumulus13@gmail.com",
    description="Custom traceback",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=install_requires,
    entry_points={
        "console_scripts": [
            "ctraceback = ctraceback.__main__:usage",
        ]
    },
    data_files=['__version__.py', 'README.md', 'LICENSE.rst'],
    license_files=["LICENSE.rst"],    
    include_package_data=True,
    python_requires=">=3",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        'License :: OSI Approved :: GNU General Public License (GPL)', 
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
)
